package comparator._2_accessor;

import java.util.Comparator;

import comparator.util.Accessor;
import comparator.util.Getter;

/**
 * {@link Comparator} mit {@link Getter}-Objekten,
 * welche ein Teil des {@link Accessor}-Interface
 * sind.
 * 
 * @param <T> Typ-Parameter, Typ der zu vergleichenden Bean
 * @author Heiner K&uuml;cker
 */
public final class AccessorComparator<T>
implements Comparator<T>
{
    /**
     * Array mit den Getter-Objekten fr 
     * die zu sortierende Bean.
     */
    private final Getter<T, ? extends Comparable<?>>[] getterArr;

    /**
     * Konstruktor.
     * 
     * @param getterArr vararg-Parameter mit {@link Getter}-Objekten fr die Propertys, nach denen sortiert werden soll
     */
    @SafeVarargs
    public AccessorComparator( 
            final Getter<T, ? extends Comparable<?>>... getterArr )
    {
        this.getterArr = getterArr;
    }

    /**
     * @see java.util.Comparator#compare
     */
    @Override
    public int compare(
            final T o1 ,
            final T o2 )
    {
        for ( final Getter<T, ? extends Comparable<?>> getter : this.getterArr )
        {
            final int getterCompareResult =
                    innerGetterCompare(
                            getter ,
                            o1 ,
                            o2 );

            if ( getterCompareResult != 0 )
            {
                return getterCompareResult;
            }
        }

        return 0;
    }
    
    /**
     * Innere Hilfsmethode zum Vergleichen einer bestimmten
     * Property der beiden bergebenen Beans.
     * <br/>
     * Dabei wird zustzlich die null-Sicherheit und die
     * Mglichkeit aufsteigend oder absteigend zu vergleichen
     * realisiert.
     * 
     * @param getter {@link Getter}-Objekt zum Zugriff auf eine bestimmte Property der zu vergleichenden Beans
     * @param o1 erste zu vergleichende Bean
     * @param o2 zweite zu vergleichende Bean
     * @return Ergebnis Vergleich als int-Wert entsprechend {@link Comparable#compareTo}
     */
    private static <B, P extends Comparable<P>> int innerGetterCompare(
            final Getter<B, ?> getter ,
            final B o1 ,
            final B o2 )
    {
        final P p1;
        final P p2;
        if ( getter instanceof InvertGetter )
            // absteigende Sortierung
        {
            final InvertGetter<B, ?> invertGetter = (InvertGetter<B, ?>) getter;
            p1 = (P) invertGetter.getterToInvert.get( o1 );
            p2 = (P) invertGetter.getterToInvert.get( o2 );
        }
        else
            // aufsteigende Sortierung
        {
            p1 = (P) getter.get( o1 );
            p2 = (P) getter.get( o2 );
        }

        final int nonInvertedResult;
        
        // Vermeiden NullPointerException beim Zugriff auf die Propertys, null gilt als kleiner
        if ( p1 == null && p2 == null )
            // beide null
        {
            // gelich
            nonInvertedResult = 0;
        }
        else if ( p1 == null )
            // links null
        {
            // links ist kleiner
            nonInvertedResult = -1;
        }
        else if ( p2 == null )
            // rechts null
        {
            // rechts ist kleiner
            nonInvertedResult = 1;
        }
        else
            // normaler Vergleich (ohne null)
        {
            nonInvertedResult =
                    p1.compareTo(
                            p2 );
        }

        if ( getter instanceof InvertGetter )
            // absteigende Sortierung
        {
            // Ergebnis invertieren
            return nonInvertedResult * -1;
        }
        else
        {
            // Ergebnis unverndert zurckgeben
            return nonInvertedResult;
        }
    }

    /**
     * Hilfsklasse zum Realisieren
     * der absteigenden Sortierung.
     * 
     * @param <B> Type-Parameter, Typ der Bean
     * @param <P> Type-Parameter, Typ der Property
     * @author Heiner K&uuml;cker
     */
    private static class InvertGetter<B, P extends Comparable<P>>
    implements Getter<B, P>
    {
        /**
         * zu invertierender {@link Getter}
         */
        private final Getter<B, P> getterToInvert;

        /**
         * Konstruktor.
         * 
         * @param getterToInvert zu invertierender {@link Getter}
         */
        private InvertGetter( 
                final Getter<B, P> getterToInvert )
        {
            this.getterToInvert = getterToInvert;
        }

        /**
         * @see Getter#get
         */
        @Override
        public P get( 
                final B bean )
        {
            throw new UnsupportedOperationException();
        }
        
    }

    /**
     * Statische Hilfsmethode zum Realisieren
     * der absteigenden Sortierung.
     * 
     * @param getterToInvert umzukehrender {@link Getter}
     * @return magisches Hll-Objekt mit umzukehrendem {@link Getter}
     */
    public static <B, P extends Comparable<P>> Getter<B, P> invert(
            final Getter<B, P> getterToInvert )
    {
        return new InvertGetter<>( getterToInvert );
    }

}
